from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.fields import ArrayField
from django.db import models
from django.utils.translation import gettext_lazy as _

from aleksis.core.mixins import ExtensibleModel


class LDAPMappingSet(ExtensibleModel):
    """Model for AlekSIS - LDAP mapping sets."""

    model = models.OneToOneField(
        ContentType, on_delete=models.CASCADE, related_name="ldap_mapping_set"
    )
    base_dn = models.CharField(max_length=255, verbose_name=_("Base DN"))
    object_classes = ArrayField(
        models.CharField(max_length=255, verbose_name=_("Object class")),
        verbose_name=_("Object classes"),
    )
    matching_mode = models.CharField(
        max_length=7,
        verbose_name=_("Matching mode"),
        choices=[("AND", _("All fields must match")), ("OR", _("Any one field must match"))],
    )

    @classmethod
    def get_for_model(cls, model):
        from django.contrib.contenttypes.models import ContentType

        content_type = ContentType.objects.get_for_model(model)
        return cls.objects.filter(model=content_type).first()

    class Meta:
        verbose_name = _("LDAP mapping set")
        verbose_name_plural = _("LDAP mapping sets")
        constraints = [
            models.UniqueConstraint(name="uniq_model_per_site", fields=["model_id", "site_id"]),
        ]


class LDAPFieldMapping(ExtensibleModel):
    """Model for AlekSIS - LDAP field mappings."""

    mapping_set = models.ForeignKey(
        LDAPMappingSet, on_delete=models.CASCADE, related_name="ldap_field_mappings"
    )
    ldap_attribute = models.CharField(max_length=255, verbose_name=_("LDAP attribute"))
    read_regex = models.CharField(max_length=255, verbose_name=_("Read regex"))
    write_template = models.CharField(max_length=255, verbose_name=_("Write template"))
    match = models.BooleanField(verbose_name=_("Use for matching"), default=False)
    include_in_rdn = models.BooleanField(verbose_name=_("Include in RDN"), default=False)

    @classmethod
    def get_for_mapping_set(cls, mapping_set):
        return cls.objects.filter(mapping_set=mapping_set)

    class Meta:
        verbose_name = _("LDAP field mapping")
        verbose_name_plural = _("LDAP field mappings")
        constraints = [
            models.UniqueConstraint(
                name="uniq_attr_per_model", fields=["mapping_set_id", "ldap_attribute"]
            ),
        ]
