"""Extension utilities for django-auth-ldap."""

import logging

logger = logging.getLogger(__name__)


class TemporaryBind:
    """LDAP conection from an LDAPUser object temporarily bound with other credentials."""

    def __init__(self, ldap_user, bind_dn, password):
        self.ldap_user = ldap_user
        self.bind_dn = bind_dn
        self.password = password

    def __enter__(self):
        if self.bind_dn is not None:
            # Bind with defined credentials and mark connection bound
            logger.info("Temporarily binding to LDAP as %s", self.bind_dn)
            self.ldap_user._bind_as(self.bind_dn, self.password, sticky=True)
        return self.ldap_user.connection

    def __exit__(self, type_, value, traceback):
        if self.bind_dn is not None:
            # Re-bind with regular credentials
            # so we do not leak connections with elevated privileges
            logger.info("Ending temporary LDAP bind as %s", self.bind_dn)
            self.ldap_user._bind()


def get_ldap_user(user):
    """Get LDAPUser for a specific Django User."""
    if not hasattr(user, "ldap_user"):
        # Add ldap_user relation to user if not available yet
        # This can happen on password reset, when the user is acted upon
        # but was never logged-in
        from django_auth_ldap.backend import LDAPBackend, _LDAPUser  # noqa

        user.ldap_user = _LDAPUser(LDAPBackend(), user=user)

    return user.ldap_user


def get_general_ldap_user():
    """Get LDAPUser with no Django User."""
    from django_auth_ldap.backend import LDAPBackend, _LDAPUser  # noqa

    # Abuse pre-configured search object as general LDAP interface
    backend = LDAPBackend()
    ldap_user = _LDAPUser(backend, "")
    return ldap_user
