import io
import logging
import re
from hashlib import shake_256

from django.apps import apps
from django.conf import settings
from django.core.exceptions import MultipleObjectsReturned, ObjectDoesNotExist
from django.core.files import File
from django.db import DataError, IntegrityError, transaction
from django.db.models import Q, fields
from django.db.models.fields.files import FileField
from django.template import Context, Template
from django.utils.text import slugify
from django.utils.translation import gettext as _

import ldap
from dynamic_preferences.types import MultipleChoicePreference, StringPreference
from ldap.modlist import addModlist, modifyModlist
from magic import Magic
from tqdm import tqdm

from aleksis.core.registries import site_preferences_registry
from aleksis.core.util.core_helpers import get_site_preferences

from .ldap import get_general_ldap_user, TemporaryBind
from .ldap_password import ldap_change_password

logger = logging.getLogger(__name__)

TQDM_DEFAULTS = {
    "disable": None,
    "unit": "obj",
    "dynamic_ncols": True,
}


def setting_name_from_field(model, field):
    """Generate a setting name from a model field."""
    name = f"additional_field_{model._meta.label_lower}_{field.name}"
    name_hash = shake_256(name.encode()).hexdigest(5)
    cleaned_name = re.sub(r"[\._]+", "_", name)
    return f"{cleaned_name}{name_hash}"


def ldap_field_to_filename(dn, fieldname):
    """Generate a reproducible filename from a DN and a field name."""
    return f"{slugify(dn)}__{slugify(fieldname)}"


def from_ldap(value, field, dn, ldap_field, instance=None):
    """Convert an LDAP value to the Python type of the target field.

    This conversion is prone to error because LDAP deliberately breaks
    standards to cope with ASN.1 limitations.
    """
    from ldapdb.models.fields import datetime_from_ldap  # noqa

    # Pre-convert DateTimeField and DateField due to ISO 8601 limitations in RFC 4517
    if isinstance(field, (fields.DateField, fields.DateTimeField)):
        # Be opportunistic, but keep old value if conversion fails
        value = datetime_from_ldap(value) or value
    elif isinstance(field, FileField) and instance is not None:
        content = File(io.BytesIO(value))

        basename = ldap_field_to_filename(dn, ldap_field)
        if ldap_field == "jpegphoto":
            extension = "jpeg"
        else:
            extension = Magic(extension=True).from_buffer(content).split("/")[0]
        name = f"{basename}.{extension}"

        # Pre-save field file instance
        fieldfile = getattr(instance, field.attname)
        fieldfile.save(name, content)

        return fieldfile

    # Finally, use field's conversion method as default
    return field.to_python(value)


def update_dynamic_preferences():
    """Auto-generate sync field settings from models."""
    from ..preferences import ldap as section_ldap  # noqa

    Person = apps.get_model("core", "Person")
    for model in (Person,):
        # Collect fields that are matchable
        for field in model.syncable_fields():
            setting_name = setting_name_from_field(model, field)

            @site_preferences_registry.register
            class _GeneratedPreference(StringPreference):
                section = section_ldap
                name = setting_name
                verbose_name = _(f"LDAP field for '{field.verbose_name}' on {model._meta.label}")
                required = False
                default = ""
                row = setting_name

            @site_preferences_registry.register
            class _GeneratedPreferenceRe(StringPreference):
                section = section_ldap
                name = setting_name + "_re"
                verbose_name = _(
                    f"Regular expression to match LDAP value for"
                    f" '{field.verbose_name}' on {model._meta.verbose_name} against"
                )
                required = False
                default = ""
                row = setting_name

            @site_preferences_registry.register
            class _GeneratedPreferenceReplace(StringPreference):
                section = section_ldap
                name = setting_name + "_replace"
                verbose_name = _(
                    f"Replacement template to apply to '{field.verbose_name}'"
                    f" on {model._meta.verbose_name}"
                )
                required = False
                default = ""
                row = setting_name

    @site_preferences_registry.register
    class LDAPMatchingFields(MultipleChoicePreference):
        section = section_ldap
        name = "matching_fields"
        default = []
        required = False
        verbose_name = _("LDAP sync matching fields")
        choices = [(field.name, field.name) for field in Person.syncable_fields()]


def apply_templates(value, patterns, templates, separator="|"):
    """Regex-replace patterns in value in order."""
    if isinstance(patterns, str):
        patterns = patterns.split(separator)
    if isinstance(templates, str):
        templates = templates.split(separator)

    for pattern, template in zip(patterns, templates):
        if not pattern or not template:
            continue

        match = re.fullmatch(pattern, value)
        if match:
            value = match.expand(template)

    return value


def get_ldap_objects(mapping_set, field_mappings, instance=None, dn=None):
    """Get results of LDAP search for a DN or a model instance with either an existing ldap_dn or configured matching fields."""
    if not instance and not dn:
        raise TypeError("Either instance or dn must be given")

    if instance and not getattr(instance, "ldap_dn", None):
        if mapping_set.matching_mode == "AND":
            search_filter = "(&"
        elif mapping_set.matching_mode == "OR":
            search_filter = "(|"

        for match in field_mappings.filter(match=True):
            template = Template(match.write_template)
            value = template.render(Context({"object": instance}))
            search_filter = f"{search_filter}({match.ldap_attribute}={value})"
        search_filter = f"{search_filter})"
    elif instance and not dn:
        dn = instance.ldap_dn

    attrlist = ["modifyTimestamp", "objectClass"]
    for field_mapping in field_mappings:
        attrlist.append(field_mapping.ldap_attribute)

    base_dn = dn if dn else mapping_set.base_dn
    search_filter = search_filter if not dn else "(objectClass=*)"

    with TemporaryBind(get_general_ldap_user(), None, None) as conn:
        results = conn.search_s(base_dn, ldap.SCOPE_SUBTREE, search_filter, attrlist=attrlist)

    return results


def read_model_from_ldap(model, dn, force=False):
    """Synchronize an LDAP object to an AlekSIS object."""
    LDAPMappingSet = apps.get_model("ldap", "LDAPMappingSet")
    LDAPFieldMapping = apps.get_model("ldap", "LDAPFieldMapping")

    mapping_set = LDAPMappingSet.get_for_model(model)
    field_mappings = LDAPFieldMapping.get_for_mapping_set(mapping_set)
    matches = {"ldap_dn": dn}
    if not mapping_set:
        return

    from aleksis.core.models import ExtensibleModel

    q = Q()
    matching_mode = mapping_set.matching_mode

    results = get_ldap_objects(mapping_set, field_mappings, dn=dn)
    if not results:
        raise ObjectDoesNotExist(f"No result for DN {dn}.")
    elif len(results) > 1:
        raise MultipleObjectsReturned("Multiple results for DN {dn}.")
    _, attrs = results[0]

    for ldap_attribute in attrs:
        field_mapping = field_mappings.filter(
            mapping_set=mapping_set, ldap_attribute__iexact=ldap_attribute
        ).first()
        if not field_mapping:
            continue

        field_matches = re.search(
            field_mapping.read_regex, attrs[ldap_attribute][0].decode("utf-8")
        )
        if not field_matches:
            raise KeyError(
                f"Field {ldap_attribute} does not match configured read regex {field_mapping.read_regex}."
            )

        model_fields = {}
        for field_name, value in field_matches.groupdict().items():
            if ExtensibleModel in model.__mro__:
                field = list(
                    filter(
                        lambda field: field.name == f"_{field_name}" or field.name == field_name,
                        model.syncable_fields(),
                    )
                )[0]
            else:
                field = model._meta.get_field(field_name)
            value = from_ldap(value, field, dn, ldap_attribute)
            model_fields[field_name] = value

            if field_mapping.match:
                add_q = Q(**{field_name: value})
                if matching_mode == "AND":
                    q = q & add_q
                elif matching_mode == "OR":
                    q = q | add_q
                else:
                    raise ValueError(f"Invalid setting for matching mode: {matching_mode}")

        matches = matches | model_fields

    try:
        instance = model.objects.get(q)
    except ObjectDoesNotExist:
        instance = model.objects.create(**matches)
    else:
        for field, value in matches.items():
            setattr(instance, field, value)
            logger.debug(f"Field {field} set to {value} for {instance}")
        instance.save()

    return instance


def write_model_to_ldap(instance):
    """Write an AlekSIS object to LDAP."""
    LDAPMappingSet = apps.get_model("ldap", "LDAPMappingSet")
    LDAPFieldMapping = apps.get_model("ldap", "LDAPFieldMapping")

    mapping_set = LDAPMappingSet.get_for_model(instance)
    field_mappings = LDAPFieldMapping.get_for_mapping_set(mapping_set)

    admin_dn = get_site_preferences()["ldap__admin_dn"]
    admin_password = get_site_preferences()["ldap__admin_password"]
    if admin_dn:
        # Use admin credentials if available
        bind_dn, password = admin_dn, admin_password
    else:
        # If not available, try using the regular LDAP auth credentials
        bind_dn, password = None, None

    results = get_ldap_objects(mapping_set, field_mappings, instance=instance)
    new_attrs = {}
    new_rdn = []

    for object_class in mapping_set.object_classes:
        if not "objectClass" in new_attrs:
            new_attrs["objectClass"] = []
        new_attrs["objectClass"].append(bytes(object_class, encoding="utf-8"))
    for field in field_mappings:
        template = Template(field.write_template)
        value = bytes(template.render(Context({"object": instance})), encoding="utf-8")
        if not field.ldap_attribute in new_attrs:
            new_attrs[field.ldap_attribute] = []
        new_attrs[field.ldap_attribute].append(value)
        if field.include_in_rdn:
            new_rdn.append(f"{field.ldap_attribute}={value.decode('utf-8')}")

    with TemporaryBind(get_general_ldap_user(), bind_dn, password) as conn:
        if not results:
            new_dn = (
                instance.ldap_dn
                if getattr(instance, "ldap_dn", None)
                else f"{'+'.join(new_rdn)},{mapping_set.base_dn}"
            )
            conn.add_s(new_dn, addModlist(new_attrs))

        elif len(results) == 1:
            dn = results[0][0]
            attrs = results[0][1]
            del attrs["modifyTimestamp"]
            new_dn = f"{'+'.join(new_rdn)},{mapping_set.base_dn}"
            if new_dn != dn:
                conn.rename_s(dn, "+".join(new_rdn))
            conn.modify_s(new_dn, modifyModlist(attrs, new_attrs))

    if hasattr(instance, "ldap_dn"):
        instance.ldap_dn = new_dn
    instance.save()

    return new_dn


@transaction.atomic
def ldap_create_user(sender, request, user, **kwargs):
    """Create a user in LDAP upon registration through allauth."""
    # Check if creation on registration is activated
    if not get_site_preferences()["ldap__user_create_on_register"]:
        return

    write_model_to_ldap(user)
    ldap_change_password(request, user, new=request.POST.get("password"))


@transaction.atomic
def ldap_sync_user_on_login(sender, user, ldap_user, **kwargs):
    """Synchronise Person meta-data and groups from ldap_user on User update."""
    # Check if sync on login is activated
    if not get_site_preferences()["ldap__person_sync_on_login"]:
        return

    Person = apps.get_model("core", "Person")

    if get_site_preferences()["ldap__enable_sync"]:
        try:
            with transaction.atomic():
                person = ldap_sync_from_user(user, ldap_user.dn, ldap_user.attrs.data)
        except Person.DoesNotExist:
            logger.warn(f"No matching person for user {user.username}")
            return
        except Person.MultipleObjectsReturned:
            logger.error(f"More than one matching person for user {user.username}")
            return
        except (DataError, IntegrityError, KeyError, ValueError) as e:
            logger.error(f"Data error while synchronising user {user.username}:\n{e}")
            return

        if get_site_preferences()["ldap__enable_group_sync"]:
            # Get groups from LDAP
            groups = ldap_user._get_groups()
            group_infos = list(groups._get_group_infos())
            group_objects = ldap_sync_from_groups(group_infos)

            # Replace linked groups of logged-in user completely
            person.member_of.set(group_objects)
            logger.info(f"Replaced group memberships of {person}")

        try:
            person.save()
        except Exception as e:
            # Exceptions here are logged only because the synchronisation is optional
            # FIXME throw warning to user instead
            logger.error(f"Could not save person {person}:\n{e}")


@transaction.atomic
def ldap_sync_from_user(user, dn, attrs):
    """Synchronise person information from a User object (with ldap_user) to Django."""
    Person = apps.get_model("core", "Person")
    create_missing = get_site_preferences()["ldap__create_missing_persons"]
    if not create_missing and not Person.objects.filter(user__username=user.username).exists():
        return
    person = read_model_from_ldap(Person, dn)

    user.save()
    person.user = user
    logger.info(f"Person {person} linked to user {user.username}")

    person.ldap_dn = dn.lower()

    person.save()
    return person


@transaction.atomic
def ldap_sync_from_groups(group_infos):
    """Synchronise group information from LDAP results to Django."""
    Group = apps.get_model("core", "Group")
    SchoolTerm = apps.get_model("core", "SchoolTerm")

    # Get current school term
    school_term = SchoolTerm.current

    # Resolve Group objects from LDAP group objects
    ldap_groups = {}
    for ldap_group in tqdm(group_infos, desc="Parsing group infos", **TQDM_DEFAULTS):
        # Skip group if one of the name fields is missing
        # FIXME Throw exceptions and catch outside
        sync_field_short_name = get_site_preferences()["ldap__group_sync_field_short_name"]
        if sync_field_short_name not in ldap_group[1]:
            logger.error(
                f"LDAP group with DN {ldap_group[0]} does not have field {sync_field_short_name}"
            )
            continue

        sync_field_name = get_site_preferences()["ldap__group_sync_field_name"]
        if sync_field_name not in ldap_group[1]:
            logger.error(
                f"LDAP group with DN {ldap_group[0]} does not have field {sync_field_name}"
            )
            continue

        # Apply regex replace from config
        short_name = apply_templates(
            ldap_group[1][get_site_preferences()["ldap__group_sync_field_short_name"]][0],
            get_site_preferences()["ldap__group_sync_field_short_name_re"],
            get_site_preferences()["ldap__group_sync_field_short_name_replace"],
        )
        name = apply_templates(
            ldap_group[1][get_site_preferences()["ldap__group_sync_field_name"]][0],
            get_site_preferences()["ldap__group_sync_field_name_re"],
            get_site_preferences()["ldap__group_sync_field_name_replace"],
        )

        # Shorten names to fit into model fields
        short_name = short_name[: Group._meta.get_field("short_name").max_length]
        name = name[: Group._meta.get_field("name").max_length]

        ldap_groups[ldap_group[0].lower()] = {"short_name": short_name, "name": name}

    all_dns = set(ldap_groups.keys())

    # First, update all existing groups with known DNs
    existing = Group.objects.filter(ldap_dn__in=all_dns, school_term=school_term).select_related(
        None
    )
    existing_dns = set([v.ldap_dn for v in existing])
    for obj in existing:
        obj.name = ldap_groups[obj.ldap_dn]["name"]
        obj.short_name = ldap_groups[obj.ldap_dn]["short_name"]
    logger.info(f"Updating {len(existing)} Django groups")
    try:
        Group.objects.bulk_update(existing, ("name", "short_name"))
    except IntegrityError as e:
        logger.error(f"Integrity error while trying to import LDAP groups:\n{e}")
    else:
        logger.debug(f"Updated {len(existing)} Django groups")

    # Second, create all groups with unknown DNs
    nonexisting_dns = all_dns - existing_dns
    nonexisting = []
    for dn in nonexisting_dns:
        nonexisting.append(
            Group(
                ldap_dn=dn,
                name=ldap_groups[dn]["name"],
                short_name=ldap_groups[dn]["short_name"],
                school_term=school_term,
            )
        )
    logger.info(f"Creating {len(nonexisting)} Django groups")
    try:
        Group.objects.bulk_create(nonexisting)
    except IntegrityError as e:
        logger.error(f"Integrity error while trying to import LDAP groups:\n{e}")
    else:
        logger.debug(f"Created {len(nonexisting)} Django groups")

    # Return all groups ever touched
    return set(existing) | set(nonexisting)


@transaction.atomic
def mass_ldap_import():
    """Add utility code for mass import from ldap."""
    from django_auth_ldap.backend import LDAPBackend, _LDAPUser  # noqa

    Person = apps.get_model("core", "Person")

    # Abuse pre-configured search object as general LDAP interface
    backend = LDAPBackend()
    connection = _LDAPUser(backend, "").connection

    # Synchronise all groups first
    if get_site_preferences()["ldap__enable_group_sync"]:
        ldap_groups = backend.settings.GROUP_SEARCH.execute(connection)
        group_objects = ldap_sync_from_groups(ldap_groups)

        # Create lookup table as cache for later code
        group_dict = {obj.ldap_dn: obj for obj in group_objects}

    # Guess LDAP username field from user filter
    uid_field = re.search(
        r"([a-zA-Z]+)=%\(user\)s", backend.settings.USER_SEARCH.searches[0].filterstr
    ).group(1)

    # Synchronise user data for all found users
    ldap_users = backend.settings.USER_SEARCH.execute(connection, {"user": "*"}, escape=False)
    for dn, attrs in tqdm(ldap_users, desc="Sync. user infos", **TQDM_DEFAULTS):
        uid = attrs[uid_field][0]

        # Prepare an empty LDAPUser object with the target username
        ldap_user = _LDAPUser(backend, username=uid)

        # Get existing or new User object and pre-populate
        user, created = backend.get_or_build_user(uid, ldap_user)
        ldap_user._user = user
        ldap_user._attrs = attrs
        ldap_user._dn = dn
        ldap_user._populate_user_from_attributes()
        user.save()

        try:
            with transaction.atomic():
                person = ldap_sync_from_user(user, dn, attrs)
        except Person.DoesNotExist:
            logger.warn(f"No matching person for user {user.username}")
            continue
        except Person.MultipleObjectsReturned:
            logger.error(f"More than one matching person for user {user.username}")
            continue
        except (DataError, IntegrityError, KeyError, ValueError) as e:
            logger.error(f"Data error while synchronising user {user.username}:\n{e}")
            continue
        else:
            logger.info(f"Successfully imported user {uid}")

    # Synchronise group memberships now
    if get_site_preferences()["ldap__enable_group_sync"]:
        member_attr = getattr(backend.settings.GROUP_TYPE, "member_attr", "memberUid")
        owner_attr = get_site_preferences()["ldap__group_sync_owner_attr"]

        for ldap_group in tqdm(
            ldap_groups,
            desc="Sync. group members",
            total=len(ldap_groups),
            **TQDM_DEFAULTS,
        ):
            dn, attrs = ldap_group

            if dn not in group_dict:
                logger.warning(f"Skip {dn} because there are no groups with this dn.")
                continue

            group = group_dict[dn]

            ldap_members = [_.lower() for _ in attrs[member_attr]] if member_attr in attrs else []

            if member_attr.lower() == "memberuid":
                members = Person.objects.filter(user__username__in=ldap_members)
            else:
                members = Person.objects.filter(ldap_dn__in=ldap_members)

            if get_site_preferences()["ldap__group_sync_owner_attr"]:
                ldap_owners = [_.lower() for _ in attrs[owner_attr]] if owner_attr in attrs else []
                if get_site_preferences()["ldap__group_sync_owner_attr_type"] == "uid":
                    owners = Person.objects.filter(user__username__in=ldap_owners)
                elif get_site_preferences()["ldap__group_sync_owner_attr_type"] == "dn":
                    owners = Person.objects.filter(ldap_dn__in=ldap_owners)

            group.members.set(members)
            if get_site_preferences()["ldap__group_sync_owner_attr"]:
                group.owners.set(owners)
            group.save()
            logger.info(f"Set group members of group {group}")

    # Synchronise primary groups
    all_persons = set(Person.objects.all())
    for person in tqdm(all_persons, desc="Sync. primary groups", **TQDM_DEFAULTS):
        person.auto_select_primary_group()
    Person.objects.bulk_update(all_persons, ("primary_group",))

    logger.info("Commiting transaction; this can take some time.")
